#!/usr/bin/env python

# rgma-cli.py
#
# Command line tool for R-GMA
#
# Mimics a SQL database command line, allowing SELECT and INSERT statements
# in standard SQL. Also provides commands to set Consumer and Producer
# properties, create Secondary Producers and query the Schema and Registry.

import rgma
import sys, cmd, string, time, os, re

class CommandLineException(Exception):
    """
    Recoverable exception thrown by the command line tool
    """

    def __init__(self, message):
               	"""
		Constructor.
		
		message - Message describing the error
		"""
		Exception.__init__(self, message)

class rgma_cli(cmd.Cmd):

    def __init__(self):
        cmd.Cmd.__init__(self)
        
        self.queryTypes = ["latest ", "history ", "continuous "]
        self.properties = ["query ", "maxage ", "timeout ", "output", "producer ", "secondaryproducer "]
        self.secondaryProperties = ["hrp ", "historyretentionperiod "]
        self.primaryProperties = self.secondaryProperties + ["lrp ", "latestretentionperiod ", "predicate "]
        self.outputFormats = ["table", "tsv", "csv"]
        
        self.queryType = rgma.QueryType.LATEST
        self.maxage = None
        self.timeout = rgma.TimeInterval(1, rgma.Units.MINUTES)
        self.outputFormat = "table"
        self.resultsFile = sys.stdout
        
        self.producer = None
        self.producerType = rgma.QueryType.CONTINUOUS
        self.producerLrp = rgma.TimeInterval(10, rgma.Units.MINUTES)
        self.producerHrp = rgma.TimeInterval(10, rgma.Units.MINUTES)
        self.producerPredicates = {}
        self.producerChanged = True
        self.producerTables = []

        self.secondaryProducer = None
        self.secondaryProducerType = rgma.QueryType.CONTINUOUS + \
                                     rgma.QueryType.LATEST
        self.secondaryProducerHrp = rgma.TimeInterval(10, rgma.Units.MINUTES)
        self.secondaryProducerPredicate = ""
        self.secondaryProducerChanged = True
        self.secondaryProducerTables = []
        
        self.directedEndpoints = None
        self.rgmaTables = []
        self.tableColumns = []
        self.terminationInterval = rgma.TimeInterval(20, rgma.Units.MINUTES)

        self.sessionHistory = []
        
        self.intro = "\nWelcome to the R-GMA virtual database for Virtual Organisations.\n" + \
                     "================================================================\n"

        serverUrl = rgma.Properties().getProperty("Consumer")[0]
        serverUrl = serverUrl[:serverUrl.rfind("/")]
        self.intro = self.intro + "\nYour local R-GMA server is:\n\n  %s\n" % serverUrl
        
        try:
            registryUrls = rgma.Properties().getProperty("Registry")
            self.registry = rgma.Registry()
            self.intro = self.intro + "\nYou are connected to the following R-GMA Registry services:\n\n"    
            for registry in registryUrls:
                self.intro = self.intro + "  %s\n" % registry
        except:
            self.registry = None
            self.intro = self.intro + "\nWARNING: failed to contact Registry service - some commands will not be available\n"

        try:
            self.schema = rgma.Schema()
            self.intro = self.intro + "\nYou are connected to the following R-GMA Schema service:\n\n  %s\n" % self.schema.url
            
        except:
            self.schema = None
            self.intro = self.intro + "\nWARNING: failed to contact Schema service - some commands will not be available\n"

            
        self.intro = self.intro + "\nType \"help\" for a list of commands.\n"

        self.prompt = "rgma> "
        self.exitcode = 0
        
    #
    # Implementations of commands
    #

    def cleanup(self):
        if self.producer:
           self.producer.close()

        if self.secondaryProducer:
           self.secondaryProducer.close()       

    def exit(self, args):
        try:
            self.cleanup()
        except (Exception), e:
            pass
        
        print "\nGoodbye\n" 
        sys.exit(0)

    def select(self, args):
        try:
            self.__query(args)
            
        except (Exception), e: 
            print e
                

    def __query(self, args):
        if len(args.strip()) == 0:
            raise CommandLineException("Usage: SELECT <columns> FROM <table> [WHERE  <predicate>] - try 'help select' for more information")
        else:
            select = "SELECT " + args
            qp = rgma.QueryProperties(self.queryType, self.maxage)
            consumer = rgma.Consumer(self.terminationInterval, \
                                     qp, \
                                     select)
            
            consumer.start(self.timeout, self.directedEndpoints) 
            first = 1
            table = 0
            numResults = 0
            warning = None
            columnWidths = None
            
            if self.outputFormat == "table":
                table = 1
            elif self.outputFormat == "tsv":
                separator = "\t"
            else:
                separator = ","

            time.sleep(0.5)
            while 1:
                results = consumer.pop(50)
                numResults += len(results)
                if results.getWarning() != None:
                    warning = results.getWarning()

                if len(results) > 0:
                    newWidths = results.getColumnWidths()
                    if columnWidths == None:
                        columnWidths = newWidths
                    else:
                        i = 0
                        for width in newWidths:
                            if columnWidths[i] < width:
                                columnWidths[i] = width
                            i += 1
                            
                    if first:
                        if table:
                            self.resultsFile.write(results.headerToString(columnWidths) + "\n")
                            first = 0
                    if table:
                        self.resultsFile.write(results.rowsToString(columnWidths) + "\n")
                    else:
                        for result in results:
                            for value in result:
                                self.resultsFile.write(value + separator)
                            self.resultsFile.write("\n")

                if results.endOfResults:
                    break

                self.resultsFile.flush()
                time.sleep(1)
                                
            if numResults == 0:
                self.resultsFile.write("No results returned\n")
            else:
                if table:
                    self.resultsFile.write(results.footerToString(columnWidths) + "\n")
                    self.resultsFile.write("%i rows\n" % numResults)

            if warning:
                self.resultsFile.write("WARNING: %s\n" % warning.getMessage())

            self.resultsFile.flush()
            consumer.close()

    def set(self, args):
        args = args.split()
        if len(args) == 0:
            raise CommandLineException("Usage SET <property> <value> - try 'help set' for more information")
        else:
            property = args[0].lower()
            if property == "query":
                self.setQuery(args[1:])
            elif property == "timeout":
                self.setTimeout(args[1:])
            elif property == "maxage":
                self.setMaxAge(args[1:])
            elif property == "output":
                self.setOutput(args[1:])
            elif property in ["producer", "p"]:
                self.setProducer(args[1:])
            elif property in ["secondaryproducer", "sp"]:
                self.setSecondaryProducer(args[1:])
            else:
                raise CommandLineException("Unrecognized property '%s' - try 'help set' for more information " % property)

    def setQuery(self, args):
        if len(args) != 1:
            raise CommandLineException("Usage SET QUERY <query type> - try 'help set' for more information")
        else:
            query = args[0].lower().strip()
            if query in rgma.QueryType.Type:
                self.queryType = rgma.QueryType.Type[query]
                print "Set query type to %s" % query
            else:
                raise CommandLineException("Unrecognized query type: '%s' - try 'help set' for usage information" % query)

    def setTimeout(self, args):
        if len(args) == 0 or len(args) > 2:
            raise CommandLineException("Usage SET TIMEOUT <value> [<units>]- try 'help set' for more information")
        timeout = self.getTimeInterval(args)
        self.timeout = timeout
        print "Set timeout to %s" % self.timeout
            
    def setMaxAge(self, args):
        if len(args) == 0 or len(args) > 2:
            raise CommandLineException("Usage SET MAXAGE <value> [<units>]- try 'help set' for more information")
        if len(args) == 1 and args[0].lower() == "none":
            self.maxage = None
            print "Set max age to none"
        else:
            maxage = self.getTimeInterval(args)
            self.maxage = maxage
            print "Set max age to %s" % self.maxage

    def setOutput(self, args):
        if len(args) != 1:
            raise CommandLineException("Usage SET OUTPUT table|tsv|csv - try 'help set' for more information")
        else:
            format = args[0].lower().strip()
            if format in self.outputFormats:
                self.outputFormat = format
                print "Set output format to '%s'" % format
            else:
                raise CommandLineException("Unrecognized output format: '%s' - try 'help set' for usage information" % format)

    def setProducer(self, args):
        """
        Handle a set producer command

        args = space-separated arguments
        """
        if len(args) == 0:
            raise CommandLineException("Usage SET PRODUCER <query types> - try 'help set' for more information")
        else:
            property = args[0].lower()
            
        if property == "predicate":
            self.setProducerPredicate(args[1:])
        elif property in ["latestretentionperiod", "lrp"]:
            self.setProducerLrp(args[1:])
        elif property in ["historyretentionperiod", "hrp"]:
            self.setProducerHrp(args[1:])
        else:
            self.setProducerQueryTypes(args)
            
    def setProducerPredicate(self, args):
        """
        Set the predicate for the primary producer

        args[0] = table name
        args[1:] = space-separated predicate
        """
        if len(args) < 2:
            raise CommandLineException("Usage SET PRODUCER PREDICATE <table> WHERE <predicate> - try 'help set' for more information")
        else:
            tableName = args[0]
            if self.producerChanged or tableName in self.producerTables:
                self.createNewPrimaryProducer()

            predicate = string.join(args[1:])

            # Check for 'none' predicate, and add WHERE if necessary
            
            if predicate.lower().strip() == "none":
                predicate = ""
            elif args[1].lower() != "where":
                predicate = "WHERE %s" % predicate
            
            self.producer.declareTable(tableName, predicate, \
                                       self.producerLrp, self.producerHrp)
            self.producerTables.append(tableName)
            
            if predicate == "":
                if tableName in self.producerPredicates:
                    del self.producerPredicates[tableName]
                print "Set producer predicate to None for table %s" % tableName
            else:
                self.producerPredicates[tableName] = predicate
                print "Set producer predicate to '%s' for table %s" % (predicate, tableName)
                
    def setProducerHrp(self, args):
        if len(args) == 0 or len(args) > 2:
            raise CommandLineException("Usage SET PRODUCER HRP <value> [<units>] - try 'help set' for more information")
        else:
            hrp = self.getTimeInterval(args)
            self.producerHrp = hrp
            print "Set producer HRP to %s" % self.producerHrp
      
    def setProducerLrp(self, args):
        if len(args) == 0 or len(args) > 2:
            raise CommandLineException("Usage SET PRODUCER LRP <value> [<units>] - try 'help set' for more information")
        else:
            lrp = self.getTimeInterval(args)
            self.producerLrp = lrp
            print "Set producer LRP to %s" % self.producerLrp
        
    def setProducerQueryTypes(self, args):
        newProducerType = rgma.QueryType.CONTINUOUS
        for query in args:
            query = query.lower()
            if query in rgma.QueryType.Type:
                newProducerType |= rgma.QueryType.Type[query]
            else:
                raise CommandLineException("Unrecognized query type: '%s' - try 'help set' for more information" % query)
        if self.producerType != newProducerType:
            self.producerType = newProducerType
            self.producerChanged = True
        print "Producer type : %s" % self.queryTypeAsString(self.producerType)
           
    def setSecondaryProducer(self, args):
        if len(args) == 0:
            raise CommandLineException("Usage SET SECONDARYPRODUCER <query types> - try 'help set' for more information")
        elif args[0].lower() in ["historyretentionperiod", "hrp"]:
            self.setSecondaryProducerHrp(args[1:])
        else:
            self.setSecondaryProducerQueryTypes(args)

    def setSecondaryProducerHrp(self, args):
        if len(args) == 0 or len(args) > 2:
            raise CommandLineException("Usage SET SECONDARYPRODUCER HRP <value> [<units>] - try 'help set' for more information")
        else:
            hrp = self.getTimeInterval(args)
            self.secondaryProducerHrp = hrp
            print "Set secondary producer HRP to %s" % self.secondaryProducerHrp
  
    def setSecondaryProducerQueryTypes(self, args):
        newProducerType = rgma.QueryType.CONTINUOUS
        for query in args:
            query = query.lower()
            if query in rgma.QueryType.Type:
                newProducerType |= rgma.QueryType.Type[query]
            else:
                raise CommandLineException("Unrecognized query type: '%s' - try 'help set' for usage information" % query)
        if self.secondaryProducerType != newProducerType:
            self.secondaryProducerType = newProducerType
            self.secondaryProducerChanged = True
        print "Secondary producer type : %s" % self.queryTypeAsString(self.secondaryProducerType)

    # new insert
    def insert(self, args):
        self.checkProducer()
        if self.producerChanged: self.createNewPrimaryProducer()
        insert = "INSERT " + args
        tableName = self.getTableNameForInsert(args)
        terminated = False
        while not terminated:
          #  print "In while loop"
            try:
                if tableName not in self.producerTables:
                    if tableName in self.producerPredicates:
                        self.producer.declareTable(tableName, self.producerPredicates[tableName], \
                                           self.producerLrp, self.producerHrp)
                    else:
                        self.producer.declareTable(tableName, "",
                                           self.producerLrp, self.producerHrp)
                        self.producerTables.append(tableName)
                
                self.producer.insert(insert)
                terminated = True
            except (rgma.UnknownResourceException), re:
                self.createNewPrimaryProducer()
                
        print "Inserted 1 row into %s" % tableName


    def show(self, args):
        args = args.split()
        
        if len(args) < 1:
            raise CommandLineException("Usage: SHOW <property> [<args>] - try 'help show' for more information")
        else:
            command = args[0].lower()
            
        if command == "tables":
            self.showTables(args[1:])
        elif command == "producers":
            self.showProducers(args[1:])
        elif command == "history":
            self.showHistory(args[1:])
        elif command in ["producer", "p"]:
            self.showProducer(args[1:])
        elif command in ["secondaryproducer", "sp"]:
            self.showSecondaryProducer(args[1:])
        elif command == "query":
            self.showQuery(args[1:])
        elif command == "timeout":
            self.showTimeout(args[1:])
        elif command == "maxage":
            self.showMaxage(args[1:])
        elif command == "properties":
            self.showProperties(args[1:])
        else:
            raise CommandLineException("Unrecognized property '%s' - try 'help show' for more information" % command)

    def showTables(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: SHOW TABLES - try 'help show' for more information")
        else:
            self.getTables()
            tuples = []
            for table in self.rgmaTables:
                tuples.append((table,))

            if len(tuples) > 0:
                rs = rgma.ResultSet(tuples, ["Table Name"])
                rs.printTable()
            else:
                print "No tables in schema"

    def showProducers(self, args):
        if len(args) != 2 or args[0].lower() != "of":
            raise CommandLineException("Usage: SHOW PRODUCERS OF <table> - try 'help show' for more information")
        else:
            table = args[1]
            if self.registry:
                producers = self.registry.getAllProducersForTable(table)
                tuples = []
                for producer in producers:
                    t = (producer.getEndpoint().getUrl(), \
                         producer.getEndpoint().getResourceId(), \
                         self.queryTypeAsString(producer.getQueryTypes()), \
                         self.booleanAsString(producer.isSecondaryProducer()))
                    tuples.append(t)

                if len(tuples) > 0:
                    rs = rgma.ResultSet(tuples, ["Endpoint", "ID", "Query types", "Secondary producer"])
                    rs.printTable()
                    print "%i producers for table %s" % (len(rs), table)
                else:
                    print "No producers for table %s" % table
            else:
                raise CommandLineException("Can't execute 'show producers' - no connection to the R-GMA Registry service is available")
                
    def showProducer(self, args):
        if len(args) == 0:
            print "Producer type: %s" % self.queryTypeAsString(self.producerType)
        else:
            property = args[0].lower()
            if property in ["hrp", "historyretentionperiod"]:
                if len(args) != 1:
                    raise CommandLineException("Usage: SHOW PRODUCER HRP - try 'help show' for more information")
                else:
                    print "Producer HRP: %s" % self.producerHrp
            elif property in ["lrp", "latestretentionperiod"]:
                if len(args) != 1:
                    raise CommandLineException("Usage: SHOW PRODUCER LRP - try 'help show' for more information")
                else:
                    print "Producer LRP: %s" % self.producerLrp
            elif property == "predicate":
                self.showProducerPredicate(args[1:])
            else:
                raise CommandLineException("No such producer property '%s' - try 'help show' for more information" % property)

    def showProducerPredicate(self, args):
        if len(args) != 1:
            raise CommandLineException("Usage: SHOW PRODUCER PREDICATE <table name> - try 'help show' for more information")
        else:
            table = args[0]
            if self.producerPredicates.has_key(table):
                predicate = self.producerPredicates.get(table)
                print "Producer predicate for '%s': %s" % (table, predicate)
            else:
                if table not in self.rgmaTables:
                    self.getTables()
                    
                if table in self.rgmaTables:
                    print "No predicate defined for table '%s'" % table
                else:
                    raise CommandLineException("No such table '%s'" % table)

    def showSecondaryProducer(self, args):
        if len(args) == 0:
            print "Secondary producer type: %s" % self.queryTypeAsString(self.secondaryProducerType)
        else:
            property = args[0].lower()
            if property in ["hrp", "historyretentionperiod"]:
                if len(args) != 1:
                    raise CommandLineException("Usage: SHOW SECONDARYPRODUCER HRP - try 'help show' for more information")
                else:
                    print "Secondary producer HRP: %s" % self.secondaryProducerHrp
            else:
                raise CommandLineException("No such secondary producer property '%s' - try 'help show' for more information" % property)

    def showQuery(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: SHOW QUERY - try 'help show' for more information")
        else:
            print "Query type: %s" % self.queryTypeAsString(self.queryType)

    def showTimeout(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: SHOW TIMEOUT - try 'help show' for more information")
        elif self.timeout:
            print "Timeout: %s" % self.timeout
        else:
            print "Timeout: None"

    def showMaxage(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: SHOW MAXAGE - try 'help show' for more information")
        elif self.maxage:
            print "Maximum age of tuples: %s" % self.maxage
        else:
            print "No maximum age of tuples"

    def showHistory(self, args):
        for command in self.sessionHistory:
            print command
        
    def showProperties(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: SHOW PROPERTIES - try 'help show' for more information")
        else:
            tuples = []
            tuples.append(("Query type", self.queryTypeAsString(self.queryType)))
        
            if self.timeout:
                timeout = self.timeout
            else:
                timeout = "None"
            tuples.append(("Query timeout", timeout))

            if self.maxage:
                maxage = self.maxage
            else:
                maxage = "None"
            tuples.append(("Max age of tuples", maxage))

            tuples.append(("Producer type", self.queryTypeAsString(self.producerType)))
            if self.producer:
                tuples.append(("Producer URL", self.producer.getEndpoint().getUrl()))
                tuples.append(("Producer ID", self.producer.getEndpoint().getResourceId()))
            else:
                tuples.append(("Producer URL", "None"))
                tuples.append(("Producer ID", "None"))
        
            for tableName in self.producerPredicates:
                tuples.append(("Producer predicate for table %s" % tableName, self.producerPredicates[tableName]))
        
            if self.producerHrp:
                tuples.append(("Producer HRP", self.producerHrp))
            else:
                tuples.append(("Producer HRP", "None"))
            if self.producerLrp:
                tuples.append(("Producer LRP", self.producerLrp))
            else:
                tuples.append(("Producer LRP", "None"))
   
            tuples.append(("Secondary producer type", self.queryTypeAsString(self.secondaryProducerType)))
            if self.secondaryProducer:
                tuples.append(("Secondary producer URL", self.secondaryProducer.getEndpoint().getUrl()))
                tuples.append(("Secondary producer ID", self.secondaryProducer.getEndpoint().getResourceId()))
            else:
                tuples.append(("Secondary producer URL", "None"))
                tuples.append(("Secondary producer ID", "None"))

            if self.producerHrp:
                tuples.append(("Secondary producer HRP", self.secondaryProducerHrp))
            else:
                tuples.append(("Secondary producer HRP", "None"))
        
            rs = rgma.ResultSet(tuples, ["Property", "Value"])
            rs.printTable()

    def use(self, args):
        
        args = args.split()
        if len(args) < 1:
            raise CommandLineException("Usage: USE PRODUCER|MEDIATOR|VDB [<args>] - try 'help use' for more information")
        else:
            command = args[0].lower()
            
        if command == "producer":
            self.useProducer(args[1:])            
        elif command == "mediator":
            self.useMediator(args[1:])
        elif command == "vdb":
            self.useVdb(args[1:])
        else:
            raise CommandLineException("Unrecognized USE command '%s' - try 'help use' for usage information" % command)

    def useVdb(self, args):
        raise CommandLineException("USE VDB not implemented yet")

    def useProducer(self, args):
        if len(args) != 2:
            raise CommandLineException("Usage: USE PRODUCER <endpoint> <resource ID> - try 'help use' for more information")
        else:
            self.directedEndpoints = [rgma.ResourceEndpoint(args[0], args[1]),]
            print "Using producer at URL %s with resource ID %s" % (args[0], args[1])

    def useMediator(self, args):
        if len(args) != 0:
            raise CommandLineException("Usage: USE MEDIATOR - try 'help use' for more information")
        else:
            self.directedEndpoints = None
            print "Using mediator to select Producers"

    def describe(self, tableName):
        if self.schema:
            table = self.schema.getTableDefinition(tableName)
            columnInfo = []
            for column in table.columns:
                columnInfo.append((column.name, rgma.TypeString(column.type, column.width),
                                   self.booleanAsString(column.primaryKey), self.booleanAsString(not column.notNull)))

            rs = rgma.ResultSet(columnInfo, ["Column name", "Type", "Primary key", "Can be NULL"])
            rs.printTable()
        else:
            raise CommandLineException("Can't execute 'describe' - no connection to the R-GMA Schema service is available")
            
    def secondaryproducer(self, args):
        self.checkSecondaryProducer()        
        if self.secondaryProducerChanged: self.createNewSecondaryProducer()
        tableName = args
        if tableName not in self.secondaryProducerTables:
            self.secondaryProducer.declareTable(tableName, "",
                                                self.secondaryProducerHrp)
            self.secondaryProducerTables.append(tableName)
        print "Secondary producer - declared table %s" % tableName

    def read(self, args):
        args = args.split()
        if len(args) != 1:
            raise CommandLineException("Usage: READ <filename> - try 'help read' for more information")

        try:
            command_file = None
            try:
                command_file = open(args[0], "r")
                commands = command_file.readlines()
            except IOError, e:
                raise CommandLineException("Could not read file '%s': %s" % (args[0], e))

            for command in commands:
                self.onecmd(self.precmd(command))
        finally:
            if command_file != None:
                command_file.close()

    def writeHistory(self, file):
        try:
            command_file = None
            try:
                command_file = open(file, "w")
                for command in self.sessionHistory:
                    command_file.write(command + "\n")
                print "Session history written to file '%s'" % file
            except IOError, e:
                raise CommandLineException("Could not write file '%s': %s" % (file, e))

        finally:
            if command_file != None:
                command_file.close()

    def writeResults(self, file):
        if file.lower() == "stdout":
            if self.resultsFile != sys.stdout:
                self.resultsFile.close()
                self.resultsFile = sys.stdout
            print "Results will be displayed on standard output"
        else:
            try:
                self.resultsFile = open(file, "w")
                print "Results will be written to file '%s'" % file
            except IOError, e:
                raise CommandLineException("Could not open file '%s': %s" % (file, e))

    def write(self, args):
        args = args.split()
        if len(args) != 2:
            raise CommandLineException("Usage: WRITE <HISTORY|RESULTS> <filename> - try 'help write' for more information")
        elif args[0].lower() == "history":
            self.writeHistory(args[1])
        elif args[0].lower() == "results":
            self.writeResults(args[1])
        else:
            raise CommandLineException("Usage: WRITE <HISTORY|RESULTS> <filename> - try 'help write' for more information")

    def clear(self, args):
        args = args.split()
        if len(args) != 1 or args[0].lower() != "history":
            raise CommandLineException("Usage: CLEAR HISTORY - try 'help clear' for more information")
        self.sessionHistory = []
        print "Session history cleared"

    def close(self, args):
        if self.producer != None:
            self.producer.close()
            self.producer = None
            self.producerChanged = True
        if self.secondaryProducer != None:
            self.secondaryProducer.close()
            self.secondaryProducer = None
            self.secondaryProducerChanged = True
            
    def create(self, args):
        if self.schema:
            statement = "CREATE " + args
            self.schema.createTable(statement, [])
        else:
            raise CommandLineException("Can't execute 'create table' - no connection to the R-GMA Schema service is available")
    
    def drop(self, args):
        
        if self.schema:
            args = args.split()
            if len(args) != 2 or args[0].lower() != "table":
                raise CommandLineException("Usage: DROP TABLE <table name> - try 'help drop' for more information")
            else:
                self.schema.dropTable(args[1])
        else:
            raise CommandLineException("Can't execute 'drop table' - no connection to the R-GMA Schema service is available")

    #
    # Helper methods
    #

    def getTables(self):
        if self.schema:
            tables = self.schema.getAllTables()
            self.rgmaTables = []
            for table in tables:
                self.rgmaTables.append(table)
        else:
            raise CommandLineException("Can't get table names - no connection to the R-GMA Schema service is available")
        
    def queryTypeAsString(self, queryType):
        str = ""
        for type in rgma.QueryType.Type.keys():
            if queryType & rgma.QueryType.Type[type]:
                str = str + type + " "
        
        return str

    def booleanAsString(self, bool):
        if bool: return "Yes"
        else: return "No"

    def getUnits(self, units):
        try:
            return rgma.Units.Unit[units.lower()]
        except KeyError:
            raise CommandLineException("Unrecognized time units: '%s' - must be seconds, minutes, hours or days" % units)

    def getTimeInterval(self, args):
        units = rgma.Units.SECONDS
        if len(args) > 2:
            raise CommandLineException("Too many arguments - time must be of the form <number> <optional units>")
        elif len(args) < 1:
            raise CommandLineException("Time value not specified")

        if len(args) >= 1:
            try:
                value = float(args[0])
            except:
                raise CommandLineException("Invalid time value '%s' - should be of the form <number> <optional units>" % args[0])
            
        if len(args) == 2:
            try:
                units = self.getUnits(args[1])
            except rgma.RGMAException:
                raise CommandLineException("Unrecognized time units: '%s' - must be seconds, minutes, hours or days" % args[1])
            
        return rgma.TimeInterval(value, units)

    def createNewPrimaryProducer(self):
        if (self.producerType == rgma.QueryType.CONTINUOUS):
            storageType = rgma.StorageType.MEMORY
        else:
            storageType = rgma.StorageType.DATABASE
            
        properties = rgma.ProducerProperties(rgma.Storage(storageType), \
                                             self.producerType)
        self.producer = rgma.PrimaryProducer(self.terminationInterval, \
                                             properties)
        self.producerTables = []
        self.producerChanged = False

    def getTableNameForInsert(self, args):
        splitArgs = args.split()
        if len(splitArgs) < 2:
            raise rgma.RGMAException("Invalid INSERT statement")
        elif splitArgs[0].lower() != "into":
            raise rgma.RGMAException("Invalid INSERT statement")
        else:
            return splitArgs[1]
 
    def createNewSecondaryProducer(self):
        properties = rgma.ProducerProperties(rgma.Storage(rgma.StorageType.DATABASE), \
                                             self.secondaryProducerType)
        self.secondaryProducer = rgma.SecondaryProducer(self.terminationInterval, \
                                                        properties)
        self.secondaryProducerChanged = False
        for table in self.secondaryProducerTables:
            self.secondaryProducer.declareTable(table, "",
                                                self.secondaryProducerHrp)

    def checkProducer(self):
       
       if self.producer is None:
           self.producerChanged = True


    def checkSecondaryProducer(self):
        
        if self.secondaryProducer is None:
            self.secondaryProducerChanged = True

    #
    # Wrapper methods for commands which handle error uniformly
    #
        
    def wrap_errors(self, function, args):
        try:
            function(args)
            self.exitcode = 0
        except (rgma.RGMAException, rgma.RemoteException,
                rgma.UnknownResourceException, rgma.RGMASecurityException), e:
            sys.stderr.write("ERROR: %s\n" % e)
            self.exitcode = 1
        except (CommandLineException), e:
            sys.stderr.write("ERROR: %s\n" % e)
            self.exitcode = 1
        except KeyboardInterrupt:
            sys.stderr.write("Interrupted\n")
            self.exitcode = 1
        except SystemExit:
            raise
        except:
            sys.stderr.write("ERROR: %s\n" % sys.exc_info()[0])
            self.exitcode = 3
            raise
        
    def do_select(self, args):
        self.wrap_errors(self.select, args)
        
    def do_set(self, args):
        self.wrap_errors(self.set, args)

    def do_insert(self, args):
        self.wrap_errors(self.insert, args)

    def do_create(self, args):
        self.wrap_errors(self.create, args)

    def do_drop(self, args):
        self.wrap_errors(self.drop, args)

    def do_show(self, args):
        self.wrap_errors(self.show, args)

    def do_use(self, args):
        self.wrap_errors(self.use, args)

    def do_describe(self, args):
        self.wrap_errors(self.describe, args)

    def do_secondaryproducer(self, args):
        self.wrap_errors(self.secondaryproducer, args)

    def do_read(self, args):
        self.wrap_errors(self.read, args)

    def do_write(self, args):
        self.wrap_errors(self.write, args)

    def do_clear(self, args):
        self.wrap_errors(self.clear, args)

    def do_close(self, args):
        self.wrap_errors(self.close, args)

    # Aliases of commands
    do_sp = do_secondaryproducer
    do_exit = exit
    do_quit = exit
    do_q = exit
    do_EOF = exit
    do_eof = exit

    #
    # Tab-completion code
    #

    def checkTables(self):
        if len(self.rgmaTables) == 0:
            try:
                self.getTables()
            except:
                pass
    
    def time_complete(self, text, args):
        if len(args) == 1:
            return self.my_complete(text, ["seconds", "minutes", "hours", "days"])
        else:
            return[]
        
    def complete_set(self, text, line, begidx, endidx):
        args = line[:begidx].split()
        num_args = len(args) - 1
        if num_args > 0:
            property = args[1].lower()
        if num_args > 1:
            subProperty = args[2].lower()
            
        if num_args == 0:
            return self.my_complete(text, self.properties)
        elif property == "query":
            return self.my_complete(text, self.queryTypes)
        elif property in ["timeout", "maxage"]:
            return self.time_complete(text, args[2:])
        elif property == "output":
            if num_args == 1:
                return self.my_complete(text, self.outputFormats)
        elif property == "producer":
            if num_args == 1:
                return self.my_complete(text, self.queryTypes + self.primaryProperties)
            elif subProperty in self.queryTypes:
                return self.my_complete(text, self.queryTypes)
            elif subProperty in ["hrp", "historyretentionperiod", "lrp", "latestretentionperiod"]:
                return self.time_complete(text, args[3:])
            elif subProperty == "predicate":
                if num_args == 3:
                    return self.my_complete(text, ["where "])
                elif num_args == 2:
                    self.checkTables()
                    return self.my_complete(text, self.rgmaTables)
            else:
                return []
        elif property == "secondaryproducer":
            if num_args == 1:
                return self.my_complete(text, self.queryTypes + self.secondaryProperties)
            elif subProperty in self.queryTypes:
                return self.my_complete(text, self.queryTypes)
            elif subProperty in ["hrp", "historyretentionperiod"]:
                return self.time_complete(text, args[3:])
        else:
            return []

    def complete_show(self, text, line, begidx, endidx):
        args = line[:begidx].split()
        num_args = len(args) - 1
        if num_args > 0:
            property = args[1].lower()
        if num_args > 1:
            subProperty = args[2].lower()

        if num_args == 0:
            return self.my_complete(text, ["tables ", "producers of ", "properties ", "history"] + self.properties)
        elif num_args == 1:
            if property == "producers":
                return self.my_complete(text, ["of"])
            elif property == "producer":
                return self.my_complete(text, self.primaryProperties)
            elif property == "secondaryproducer":
                return self.my_complete(text, self.secondaryProperties)
        elif num_args == 2:
            self.checkTables()
            if property == "producers" and subProperty == "of":
                return self.my_complete(text, self.rgmaTables)
            elif property == "producer" and subProperty == "predicate":
                return self.my_complete(text, self.rgmaTables)
        else:
            return []
          
    def complete_use(self, text, line, begidx, endidx):
        existing_commands = line[:begidx].split()
        if len(existing_commands) == 1:
            return self.my_complete(text, ["producer ", "mediator "])
        else:
            return []
        
    def complete_select(self, text, line, begidx, endidx):
        line = line[:begidx]
        afterSelect = re.compile(r"select.*", re.IGNORECASE).match(line)
        afterFrom = re.compile(r"select\s+.*from.*", re.IGNORECASE).match(line)
        afterWhere = re.compile(r"select\s+.*from\s+.*where.*", re.IGNORECASE).match(line)
        
        if afterWhere:
            return self.my_complete(text, self.tableColumns + ["order by ", ""])
        elif afterFrom:
            self.checkTables()
            return self.my_complete(text, self.rgmaTables + ["where ", "order by "])
        elif afterSelect:
            return self.my_complete(text, self.tableColumns + ["* ", "from "])
        else:
            return self.my_complete(text, [])

    def complete_insert(self, text, line, begidx, endidx):
        existing_commands = line[:begidx].split()
        if len(existing_commands) == 1:
            return self.my_complete(text, ["into"])
        if existing_commands[len(existing_commands) - 1].lower() == "into":
            self.checkTables()
            return self.my_complete(text, self.rgmaTables)
    
    def complete_describe(self, text, line, begidx, endidx):
        existing_commands = line[:begidx].split()
        if len(existing_commands) == 1:
            self.checkTables()
            return self.my_complete(text, self.rgmaTables)
        else:
            return []

    def complete_secondaryproducer(self, text, line, begidx, endidx):
        self.checkTables()
        return self.my_complete(text, self.rgmaTables)

    def complete_write(self, text, line, begidx, endidx):
        return self.my_complete(text, ["history ", "results "])

    def complete_clear(self, text, line, begidx, endidx):
        return self.my_complete(text, ["history"])

    def complete_close(self, text, line, begidx, endidx):
        return self.my_complete(text, ["producers"])
        
    # Aliases
    
    complete_sp = complete_secondaryproducer

    def my_complete(self, text, known_commands):
        matches = []
#        text = text.lower()
        for command in known_commands:
            if command.startswith(text):
                matches.append(command)

        return matches

    #
    # Help documentation
    #
    
    def help_examples(self):
        print
        print "Example commands"
        print "================"
        print
        print "set query latest"
        print "select Name, Latitude, Longitude from Site"
        print
        print "  -- Basic SELECT query on Site table"
        print
        print "set query continuous"
        print "set maxage 1 days"
        print "set timeout 10 seconds"
        print "select * from Site"
        print
        print "  -- Continuous + History SELECT query on Site table"
        print
        print "set producer continuous"
        print "insert into userTable values ('id', 'string', 1.2, 3)"
        print "set query continuous"
        print "set maxage 1 minutes"
        print "set timeout 5 seconds"
        print "select * from userTable where userId='id'"
        print
        print "  -- INSERT, then SELECT using Continuous + History query"
        print
        print "set secondaryproducer latest"
        print "secondaryproducer userTable"
        print "set producer continuous"
        print "insert into userTable values ('id2', 'string', 1.2, 3)"
        print "set query latest"
        print "select * from userTable where userId='id2'"
        print
        print "  -- INSERT, then SELECT using Secondary Producer to support"
        print "     the latest query"
        print
        print "show tables"
        print
        print "  -- Show names of all tables defined in the Schema"
        print
        print "describe userTable"
        print
        print "  -- Print table description"
        print
        print "set producer continuous"
        print "insert into userTable values ('id3', 'string', 1.2, 3)"
        print "show producers of userTable"
        print
        print "  -- Show list of producers for a particular table"
        
        
    def help_select(self):
        print
        print "select <columns> from <table> [where <predicate>] [order by <column>]"
        print "====================================================================="
        print "Query R-GMA for information contained in <columns> of <table>"
        print "which satisfies <predicate>, ordering the results according to"
        print "the values in <column>. The type of query can be changed"
        print "using the 'set query' command."
        print
        print "Note that 'order by' is not supported for continuous queries"
        print "since in this case the tuples are returned in the order they"
        print "are inserted."
       
    def help_set(self):
        print
        print "set <property> <value>"
        print "======================"
        print "Set properties for queries and inserts"
        print
        print "  set query latest|history|continuous"
        print "    - Set the query type for SELECT statements"
        print "  set timeout <timeout>"
        print "    - Set the timeout for queries in seconds"
        print "  set timeout <timeout> <units>"
        print "    - Set the timeout for queries in <units>"
        print "      <units> may be seconds, minutes, hours or days"
        print "  set maxage none"
        print "    - Disable maximum age restriction for tuples"
        print "  set maxage <time>"
        print "    - Set the maximum age of tuples to return in seconds"
        print "  set maxage <time> <units>"
        print "    - Set the maximum age of tuples to return in <units>"
        print "      <units> may be seconds, minutes, hours or days"
        print "  set output table|tsv|csv"
        print "    - Set the output format for results. 'table' formats the"
        print "      results in  a table, tsv outputs tab-separated results"
        print "      and csv outputs comma-separated results"
        print "  set producer [latest] [history] [continuous]"
        print "    - Set supported query types for the producer"
        print "      May be any combination of latest, history and continuous"
        print "      N.B. All producers support continuous queries so specifying"
        print "      continuous currently has no effect"
        print "  set producer predicate <table> [WHERE] <predicate>"
        print "    - Set the producer predicate for <table>. The WHERE keyword is optional."
        print "      All tuples inserted into <table> must be consistent with the predicate."
        print "  set producer latestretentionperiod|lrp <time value> [<units>]"
        print "    - Set the latest retention period for tuples published by the"
        print "      producer."
        print "  set producer historyretentionperiod|hrp <time value> [<units>]"
        print "    - Set the history retention period for the producer. If the producer"
        print "      does not support history queries this command has no effect."
        print "  set secondaryproducer [latest] [history] [continuous]"
        print "    - Set supported query types for the secondary producer"
        print "      May be any combination of latest, history and continuous"
        print "      N.B. All producers support continuous queries so specifying"
        print "      continuous currently has no effect" 
        print "  set secondaryproducer historyretentionperiod|hrp <time value> [<units>]"
        print "    - Set the history retention period for the secondary producer. If the"
        print "      secondary producer does not support history queries this command has"
        print "      no effect."
        
    def help_insert(self):
        print
        print "insert into <table> [(<col1>, <col2>...)] values (<value1>, <value2>...)"
        print "=============================================="
        print "Inserts the specified values into the specified columns of <table> using the"
        print "primary producer. If the columns are not specified, the default table column"
        print "order is assumed. The query types supported by the primary producer can be"
        print "changed using the 'set producer' command"

    def help_show(self):
        print
        print "show <property> <arguments>"
        print "==========================="
        print "Show the value of a property or properties"
        print
        print "  show tables"
        print "    - Show a list of tables defined in the Schema"
        print "  show producers of <table>"
        print "    - Show a list of endpoints and resource IDs of producers that"
        print "      publish to <table>"
        print "  show timeout"
        print "    - Show current query timeout"
        print "  show maxage"
        print "    - Show current maximum age of tuples to return"
        print "  show query"
        print "    - Show current query type"
        print "  show producer"
        print "    - Show current producer type"
        print "  show producer predicate <table>"
        print "    - Show the producer predicate for <table>"
        print "  show producer lrp|latestretentionperiod"
        print "    - Show current producer latest retention period"
        print "  show producer hrp|historyretentionperiod"
        print "    - Show current producer history retention period"
        print "  show secondaryproducer"
        print "    - Show current secondary producer type"
        print "  show secondaryproducer hrp|historyretentionperiod"
        print "    - Show current secondary producer history retention period"
        print "  show history"
        print "    - Each command issued is added to the session history"
        print "      (whether it is successful or not). The session history"
        print "      can be emptied using the 'clear history' command and"
        print "      written to a file using the 'write history' command"
        print "  show properties"
        print "    - Show summary of all properties"

    def help_describe(self):
        print
        print "describe <table>"
        print "================"
        print "Show column names and types for <table>"
        
    def help_exit(self):
        print
        print "exit, quit, q"
        print "============="
        print "Exit the R-GMA command line tool"
              
    def help_secondaryproducer(self):
        print
        print "secondaryproducer <table>"
        print "========================="
        print "Declare <table> to be consumed and republished by the"
        print "secondary producer. The types of query supported by the"
        print "secondary producer can be changed using the"
        print "'set secondaryproducer' command"

    def help_use(self):
        print
        print "use producer <url> <id>"
        print "======================="
        print "Force all subsequent SELECT queries to use only the specified"
        print "producer. Only one producer may be specified."
        print "Note that if the producer is invalid or cannot answer the query"
        print "type used, no results will be returned"
        print
        print "use mediator"
        print "============"
        print "Use the R-GMA mediator to select producers to answer queries."
        print "This overrides any previous 'use producer' commands."
        print
        print "use vdb <name>"
        print "============="
        print "Not implemented yet"

    def help_read(self):
        print
        print "read <filename>"
        print "==============="
        print "Read in the specified file and execute the commands contained in it."
        print "One command must be specified per line."

    def help_write(self):
        print
        print "write history <filename>"
        print "========================"
        print
        print "Write the session command history to a file."
        print
        print "Each command issued is added to the session history"
        print "(whether it is successful or not). The session history"
        print "can be displayed using the 'show history' command and "
        print "emptied using the 'clear history' command"
        print
        print "write results <filename>"
        print "========================"
        print
        print "Write query results to a file."
        print
        print "Results of SELECT queries will be written to <filename>"
        print "If <filename> is given as 'stdout', standard output will"
        print "be used, i.e. results written to the screen. This is the"
        print "default."
        
    def help_clear(self):
        print
        print "clear history"
        print "============="
        print
        print "Reset the session command history."
        print
        print "Each command issued is added to the session history"
        print "(whether it is successful or not). The session history"
        print "can be displayed using the 'show history' command and written"
        print "to a file using the 'write history' command"

    def help_close(self):
        print
        print "close producers"
        print "==============="
        print
        print "Close the current Primary and Secondary producers"
        print
        print "This may be used prior to exit to ensure that no redundant"
        print "producers are left in existance. It may also be used to ensure"
        print "that new producers are generated for subsequent 'insert' and"
        print "'secondaryproducer' statements"
        
    def help_create(self):
        print
        print "create table <table name> (col1 type1 [PRIMARY KEY], col2...)"
        print "============================================================="
        print
        print "Create a table in the R-GMA virtual database schema"

    def help_drop(self):
        print
        print "drop table <table name>"
        print "======================="
        print
        print "Delete a table from the R-GMA virtual database schema"

    def help_help(self):
        print
        print "help <command>"
        print "============"
        print
        print "Print help on the specified command"
        print
        print "Use 'help' for a list of documented commands"
        print "Use 'help examples' for a list of example commands"

    def help_EOF(self):
        print
        print "EOF"
        print "==="
        print
        print "Send end-of-file signal. This will exit the R-GMA command"
        print "line processor"
        
    # Aliases
    
    help_sp = help_secondaryproducer
    help_quit = help_exit
    help_q = help_exit
    help_eof = help_EOF

    #
    # Generic command processing functions
    #
    
    def precmd(self, command):
        """
        Strip everything after the '#' character - this represents
        a comment.
        
        Also strip trailing ';' from command to avoid tripping up people
        who are used to MySQL. Also, lowercase the first word as the first
        step towards case insensitivity.
        """
        processed_command = command.strip().split('#', 1)[0].rstrip(";")
        
        splitCommand = processed_command.split(" ", 1)
        splitCommand[0] = splitCommand[0].lower()
        if splitCommand[0] == "help":
            if len(splitCommand) > 1:
                splitCommand[1] = splitCommand[1].lower()
        elif splitCommand[0] not in ["clear", "write"]:
            if len(splitCommand) == 1:
                self.sessionHistory.append(command)
            elif splitCommand[0] != "show" or splitCommand[1] != "history":
                self.sessionHistory.append(command)

        return string.join(splitCommand)
    
    def default(self, line):
        """
        Unrecognized command
        """
        sys.stderr.write("ERROR: Unknown command: '%s' -try 'help' for a list of commands\n" % line)
        self.exitcode = 1
        
    def emptyline(self):
        return

#
# Startup and command-line argument handling code
#
    
def showHelp():
    print
    print "R-GMA command line tool"
    print "======================="
    print
    print "Provides command-line access to the R-GMA information and"
    print "monitoring system"
    print
    print "Usage: rgma [-h, --help] [-v, version] [-c <command>] [-f <file>]"
    print
    print "  -h, --help    - Display usage information"
    print "  -v, --version - Display version"
    print "  -c <command>  - Execute <command> then exit. If specified"
    print "                  more than once, execute commands in the order"
    print "                  they appear, then exit"
    print "  -f <file>     - Execute commands in <file>, in batch"
    print
    
def showVersion():
    print
    print "R-GMA command line tool version @VERSION@"
    print
    
if __name__ == "__main__":

   command_next = 0
   commands = []
   file_next = 0
   batch_file = None
   
   for arg in sys.argv[1:]:
                          
       if command_next:
           commands.append(arg)
           command_next = 0
       elif file_next:
           batch_file = arg
           file_next = 0
       elif arg in ["--help", "-h"]:
           showHelp()
           sys.exit(0)
       elif arg in ["--version", "-v"]:
           showVersion()
           sys.exit(0)
       elif arg == "-c":
           command_next = 1
       elif arg == "-f":
           file_next = 1
       else:
           sys.stderr.write("\nERROR - unrecognized command line option %s\n" % arg)
           sys.stderr.write("Try rgma --help for usage information\n")
           sys.exit(1)
                       
   if command_next:
       sys.stderr.write("\nERROR: -c option used but no command specified\n")
       sys.stderr.write("Try rgma --help for usage information\n")
       sys.exit(1)

   if file_next:
       sys.stderr.write("\nERROR: -f option used but no file specified\n")
       sys.stderr.write("Try rgma --help for usage information\n")
       sys.exit(1)
       
   if len(commands) > 0 and batch_file:
       sys.stderr.write("\nERROR: cannot specify -c and -f together\n")
       sys.stderr.write("Try rgma --help for usage information\n")
       sys.exit(1)
       
   rgma_cli = rgma_cli()
   if len(commands) > 0:
       for command in commands:
           rgma_cli.onecmd(rgma_cli.precmd(command))

       rgma_cli.cleanup();
       sys.exit(rgma_cli.exitcode)
   elif batch_file:
       rgma_cli.do_read(batch_file)
       exitcode = rgma_cli.exitcode
       rgma_cli.cleanup();
       sys.exit(exitcode)
   else:
       try:
           rgma_cli.cmdloop()
       except KeyboardInterrupt:
           rgma_cli.exit(None)
